(function (root, moduleName, factory) {
    'use strict';
    if (typeof define === 'function' && define.amd) {
        define([], factory);
    } else if (typeof module !== 'undefined' && module.exports) {
        module.exports = factory.call(root);
    } else {
        root[moduleName] = factory.call(root);
    }
}
    (typeof(window) !== 'undefined' ? window : typeof exports !== 'undefined' ? exports : typeof global !== 'undefined' ? global : this, 'CSSPlugin', function (undefined) {
        'use strict';

        // Some variables to be working correctly

        // Get prefix
        var doc = window.document;
        var HYPEN_PREFIX = null;
        var Prefixes = ["Webkit", "Moz", "Ms", "O"];
        var Property = "Transform";
        var el,
        s;
        if (!doc || !doc.createElement) {
            HYPEN_PREFIX = 'Webkit'; // It should be Webkit/V8 as it's Node.js
        } else {
            el = document.createElement('div');
            s = el.style;
            if (Property in s) {
                HYPEN_PREFIX = ''; // Modern prefix-free browsers
            } else {
                for (var i = 0, len = Prefixes.length; i < len; i++) {
                    var pref = Prefixes[i];
                    if (pref in s) {
                        HYPEN_PREFIX = pref;
                        break;
                    }
                }
            }
        }
		var CSS_PREFIX = HYPEN_PREFIX !== '' ? "-" + HYPEN_PREFIX.toLowerCase() + "-" : '';

		// Prefix
		var _prefix = function (property) {
			var propPrefixed = String(HYPEN_PREFIX + property[0].toUpperCase() + property.substr(1));
			return propPrefixed in s ? propPrefixed : property;
		};

        // Check SVG Elem
        var hasSVGSupport = window.SVGElement;
        var isSVGElement = function (node) {
            return hasSVGSupport && node instanceof SVGElement;
        };

        // Check 3D support for CSS3 Transform
        var IS3D = _prefix('perspective') in s;

        // Free memory leak
        el = null;
        s = null;

        // Performance Optimization
        var GlobalRoundProps = {
            left: true,
            top: true,
            right: true,
            bottom: true
        };

        var TransformProps = {
            x: true,
            y: true,
            z: true,
            translate: true,
            translate3d: true,
            scale: true,
            scaleX: true,
            scaleY: true,
            scaleZ: true,
            scale3d: true,
            rotateX: true,
            rotateY: true,
            rotateZ: true,
            rotate: true,
            skew: true,
            skewX: true,
            skewY: true,
            transformPerspective: true
        };

		var ExceptedProperty = {
			transformOrder: true,
			clearProps: true,
			roundProps: true,
			decimalProps: true,
			transformOrigin: true
		};

		var AttrProperty = {
			d: true,
			points: true,
			x1: true,
			y1: true,
			x2: true,
			y2: true,
			r: true,
			rx: true,
			ry: true,
			cx: true,
			cy: true,
			stopColor: true
		};

		var ScrollProperty = {
			scrollTop: true,
			scrollLeft: true
		};

        function def(v, d, u) {
            return typeof v === 'number' && u ? (v !== undefined ? v : d) + u : v !== undefined ? v : d;
        }

        var defaultTransformOrder = ["transformPerspective", "x", "y", "z", "scaleX", "scaleY", "scaleY", "scale", "rotateX", "rotateY", "rotateZ", "rotate", "skewX", "skewY", "skew"];

        var transformProp = _prefix('transform');

        var CSSPlugin = function CSSPlugin(Tween, Start, End) {

            var node = Tween.node;

			if (!node) {
				return false;
			}

			this.node = node;
            this._style = node.style;
            var transformOrder = End.transformOrder || defaultTransformOrder;
            var clearProps = End.clearProps;
            this.roundProps = End.roundProps;
            var transformOrigin = End.transformOrigin,
            isSVGNode = isSVGElement(node),
            is3DNode = IS3D && !isSVGNode,
            orderLen = transformOrder.length;

            if (transformOrder.length < defaultTransformOrder.length) {
                transformOrder = defaultTransformOrder.reduce(function (p, c) {
                        if (p.indexOf(c) === -1) {
                            p.push(c);
                        }
                        return p;
                    }, transformOrder);
            }
            this.transformOrder = transformOrder;

            var transformOriginFunc;
            var bbox,
            originTweenX = 0,
            originTweenY = 0,
            setTransformDOM = false;

            var pxsfx = isSVGNode ? 0 : 'px';
            var degsfx = isSVGNode ? 0 : 'deg';
            var smoothOriginPrepend = '',
            smoothOriginAppend = '';
            var _style = node.style;

			if (!transformOrigin && isSVGNode) {
				transformOrigin = "0% 0%";
			}
            if (transformOrigin) {
                if (isSVGNode) {
                    transformOriginFunc = function (origin, i) {
                        if (origin.indexOf("px") !== -1) {
                            return bbox[i === 0 ? "x" : "y"] + parseFloat(origin);

                        } else if (origin.indexOf("%") !== -1) {
                            return bbox[i === 0 ? "x" : "y"] + (bbox[i === 0 ? "width" : "height"] * (parseFloat(origin) / 100))
                        }
                        return parseFloat(origin);
                    };
                    bbox = node.getBBox();
                    transformOrigin = transformOrigin.split(" ").map(transformOriginFunc);
                    originTweenX = transformOrigin[0];
                    originTweenY = transformOrigin[1];
                    smoothOriginPrepend = 'translate(' + originTweenX + ',' + originTweenY + ') ';
                    smoothOriginAppend = 'translate(' + -originTweenX + ',' + -originTweenY + ') ';
                } else {
                    transformOriginFunc = function (origin, i) {
                        if (origin.indexOf("%") !== -1) {
                            return (node[i === 0 ? "offsetWidth" : "offsetHeight"] * (parseFloat(origin) / 100));
                        }
                        return parseFloat(origin);
                    };
                    var orig = transformOrigin.split(" ").map(transformOriginFunc);
                    node.style.transformOrigin = transformOrigin;
                    originTweenX = orig[0];
                    originTweenY = orig[1];
                    if (is3DNode) {
                        smoothOriginPrepend = 'translate3d(' + originTweenX + pxsfx + ',' + originTweenY + pxsfx + ', 0px) ';
                    } else {
                        smoothOriginPrepend = 'translate(' + originTweenX + pxsfx + ',' + originTweenY + pxsfx + ') ';
                    }
                }
                this.transformOrigin = transformOrigin;

            }
            this.is3DNode = is3DNode;
            this.isSVGNode = isSVGNode;
            this.pxsfx = pxsfx;
            this.degsfx = degsfx;
            this.setTransformDOM = setTransformDOM;
            this.smoothOriginPrepend = smoothOriginPrepend;
            this.smoothOriginAppend = smoothOriginAppend;
            this.clearProps = clearProps;

			for (var prop in End) {
				var prefProp = _prefix(prop);
				if (prop !== prefProp) {
					var vs = Start[prop], ve = End[prop];
					Start[prop] = null;
					End[prop] = null;
					Start[prefProp] = vs;
					End[prefProp] = ve;
					delete Start[prop];
					delete End[prop];
				)
			}
			return this;
        };
        var p = CSSPlugin.prototype = {
            setOrigin: function (origin) {
                if (isSVGNode) {
                    var oldX = transformOrigin[0];
                    var oldY = transformOrigin[1];
                    var orig = origin.split(" ").map(transformOriginFunc);
                    var newX = orig[0];
                    var newY = orig[1];
                    var diffX = newX - oldX;
                    var diffY = newY - oldY;
                    originTweenX = newX;
                    originTweenY = newY;
                    transformOrigin[0] += diffX;
                    transformOrigin[1] += diffY;
                    var xOrig = transformOrigin[0],
                    yOrig = transformOrigin[1];
                    originTweenX += Math.abs(xOrig - oldX);
                    originTweenY += Math.abs(yOrig - oldY);
                    smoothOriginPrepend = 'translate(' + originTweenX + ',' + originTweenY + ')';
                    smoothOriginAppend = 'translate(' + -xOrig + ',' + -yOrig + ')';
                } else {
                    var orig = origin.split(" ").map(transformOriginFunc);
                    originTweenX += orig[0];
                    originTweenY += orig[1];
                    transformOrigin = origin;
                    setTransformDOM = false;
                    if (is3DNode) {
                        smoothOriginPrepend = 'translate3d( ' + originTweenX + pxsfx + ',' + originTweenY + pxsfx + ',0px)';
                    } else {
                        smoothOriginPrepend = 'translate(' + originTweenX + pxsfx + ',' + originTweenY + pxsfx + ')';
                    }
                    this.smoothOriginPrepend = smoothOriginPrepend;
                    this.smoothOriginAppend = smoothOriginAppend;
                }
				return this;
            },
            update: function (style, t) {

                var roundProps = this.roundProps;
                var transformOrigin = this.transformOrigin;
                var clearProps = this.clearProps;
				var node = this.node;
                var _style = this._style;
                var transformOrder = this.transformOrder;
                var is3DNode = this.is3DNode;
                var isSVGNode = this.isSVGNode;
                var setTransformDOM = this.setTransformDOM;
                var pxsfx = this.pxsfx;
                var degsfx = this.degsfx;
                var smoothOriginPrepend = this.smoothOriginPrepend;
                var smoothOriginAppend = this.smoothOriginAppend;
				var i = 0;
                var orderLen = transformOrder.length;
                var transform;
                var translateSet = null;

                for (var p in style) {
                    if (typeof style[p] === 'number') {
                        if (roundProps && roundProps.indexOf(p) !== -1 || GlobalRoundProps[p]) {
                            style[p] = style[p] | 0;
                        }
                    }
                    if (TransformProps[p] || ExceptedProperty[p]) {
						continue;
					} else if (ScrollProperty[p]) {
						node[p] = style[p];
					} else if (AttrProperty[p]) {
						node.setAttribute(p, style[p]);
					} else {
                        _style[p] = style[p];
                    }
                }

                for (; i < orderLen; i++) {
                    var p = transformOrder[i];
                    if (style[p] !== undefined) {
                        transform = transform || '';
                        if (p === 'x' || p === 'y' || p === 'z') {
                            if (!translateSet) {
                                transform += is3DNode ? 'translate3d(' + def(style.x, 0, 'px') + ',' + def(style.y, 0, 'px') + ',' + def(style.z, 0, 'px') + ') ' : 'translate(' + def(style.x, 0, pxsfx) + ',' + def(style.y, 0, pxsfx) + ') ';
                                translateSet = true;
                            }
                        } else if (p === 'scale' || p === 'scaleX' || p === 'scaleY' || p === 'scaleZ') {
                            transform += is3DNode ? 'scale3d(' + def(style.scaleX, style.scale !== undefined ? style.scale : 1, 0) + ',' + def(style.scaleY, style.scale !== undefined ? style.scale : 1, 0) + ',' + def(style.scaleZ, style.scale !== undefined ? style.scale : 1, 0) + ') ' : 'scale(' + def(style.scaleX, style.scale !== undefined ? style.scale : 1, 0) + ',' + def(style.scaleY, style.scale !== undefined ? style.scale : 1, 0) + ') ';
                        } else if (p === 'transformPerspective') {
                            transform += 'perspective(' + style[p] + ') ';
                        } else if (p === 'rotate' || p === 'rotateZ') {
                            transform += p + '(' + def(style[p], 0, degsfx) + ') ';
                        } else {
                            transform += p + '(' + style[p] + ') ';
                        }
                    }
                }

                if (transform) {
                    if (isSVGNode) {
                        node.setAttribute('transform', smoothOriginPrepend + transform + smoothOriginAppend);
                    } else {
                        if (!setTransformDOM && transformOrigin) {
                            _style.transformOrigin = transformOrigin;
                            this.setTransformDOM = true;
                        }
                        _style[transformProp] = smoothOriginPrepend + transform;
                    }
                    transform = null;
                }

                if (t === 1 && clearProps) {
                    for (var i = 0, len = clearProps.length; i < len; i++) {
                        var p = clearProps[i];
                        if (isSVGNode && node.getAttribute(p)) {
                            node.removeAttribute(p);
                        } else {
                            _style[p] = sp;
                        }
                    }
                }
            }
        }

		if (this.TWEEN !== undefined) {
			TWEEN.Tween.prototype.Renderer = CSSPlugin;
		}

        return CSSPlugin;

    }));
